"""
for drawing networks/graphs
"""
import re
from pathlib import Path

import graphviz
import lxml.etree as ET
import networkx as nx
from matplotlib import pyplot as plt

from amilib.ami_html import HtmlLib, logger
from amilib.file_lib import FileLib
from test.resources import Resources


class AmiGraph:
    """
    grahs can be nested?
    """

    def __init__(self):
        self.subgraphs = []
        self.graph = None
        self.html_elem = None

    def create_subgraph(self):
        subgraph = AmiGraph()
        self.subgraphs.append(subgraph)

    @classmethod
    def create_nested_uls_from_nested_divs(cls, html_elem):
        """
        treat all divs and their immediate childresn as nodes with
        edges parent->child
        """

        def _iterate_children(elem, new_html, xpath):
            children = elem.xpath(xpath)
            if (nchild := len(children)) == 0:
                return None
            ul = ET.SubElement(new_html, "ul")
            p = ET.SubElement(ul, "p")
            p.text = f"{nchild}.."
            for child in children:
                # print(f"{child.tag} => ({child.get('class')})")
                li = ET.SubElement(ul, "li")
                h = child.xpath("h1 | h2 | h3 | h4 | a")
                if h:
                    h0 = h[0]
                    li.text = h0.text
                    if h0.text is None:
                        print(f"no text for {child.tag}@({child.attrib.get('class')})")
                    else:
                        print(f" ... {h0.text}")
                else:
                    print(f"no title for {child.attrib.get('class')}")
                _iterate_children(child, li, xpath)

        body = HtmlLib.get_body(html_elem)
        new_body, new_html = HtmlLib.create_html_with_body()
        _iterate_children(body, new_body, xpath="div | header | section")
        return new_html


    def create_network_from_toc(self, toc_path):
        """
        Creates the graph from a nested list which is present in toc.html.
        This is a function which is called in 'test_toc_workflow'.

        """
        self.graph = nx.DiGraph()
        html_elem = HtmlLib.parse_html(toc_path)
        root_ul = html_elem.find(".//ul")
        root_ul.attrib["id"] = "root"
        root_ul.attrib["title"] = str(toc_path)
        return self.graph


        def add_ids_to_nodes(html_elem):
            li_nodes = html_elem.findall(".//li")
            for i, li_node in enumerate(li_nodes):
                li_node.attrib["id"] = f"li_{i}"
            print(f"li nodes {len(li_nodes)}")
            ul_nodes = html_elem.findall(".//ul")
            for i, ul_node in enumerate(ul_nodes):
                ul_node.attrib["id"] = f"ul_{i}"
            print(f"ul nodes {len(ul_nodes)}")


    def add_nodes_to_network(self, ul):
        print(f"add_nodes {ul} {ul.attrib.get('id')}")
        ul_id = ul.attrib.get("id")
        for li in ul.findall("./li"):
            li_id = li.attrib.get("id")
            print(f"{li_id}")
            text = li.text.strip() if li.text else li_id
            # Add node to network
            self.graph.add_node(li_id)
            if ul is not None:
                # Adding edge to network
                self.graph.add_edge(ul_id, li_id)
            nested_ul = li.find("./ul")
            if nested_ul is not None:
                self.add_nodes_to_network(nested_ul)

        self.add_nodes_to_network(root_ul)
        return self.graph

    @classmethod
    def visualize_graph_with_networkx(cls, graph, title="Graph", wg=None):
        """
        Visualizes the graph using a fallback layout and saves it as an image.

        Parameters:
            graph: The NetworkX graph to visualize.
            title: Title of the graph.
            wg: Working group name (used for saving files).
            Used layout as spring_layout.
        """

        assert graph is not None
        pos = nx.spring_layout(graph, k=2.0)

        """
        Attributes to draw the network
        """
        plt.figure(figsize=(12, 8))
        nx.draw(
            graph,
            pos,
            with_labels=True,
            node_size=2000,
            node_color="skyblue",
            font_size=10,
            font_weight="bold",
            edge_color="gray",
            arrows=True,
            width=1.5
        )
        plt.title(title)

        """
        Saving the network as png.
        """
        if wg:
            image_path = Path(Resources.TEMP_DIR, "ipcc", wg, f"{wg}_nxgraph.png")
            plt.savefig(image_path, format="PNG", bbox_inches="tight")
            print(f"Graph image saved as {image_path}")
        """
        To Save the graph as a DOT file
        if wg:
            dot_path = Path(Resources.TEMP_DIR, "ipcc", wg, f"{wg}_graph.dot")
            nx.nx_pydot.write_dot(graph, dot_path)
            print(f"Graph saved as DOT file: {dot_path}")

        To Save the graph as a GML file
        if wg:
            gml_path = Path(Resources.TEMP_DIR, "ipcc", wg, f"{wg}_graph.gml")
            nx.write_gml(graph, gml_path)
            print(f"Graph saved as GML file: {gml_path}")

        To Show the graph (optional)
        plt.show()
        """

    @classmethod
    def create_and_display_chapter_toc_network(cls, wg, chapter, infile, gv_output, maxlev=4, engine='fdp'):
        """
        :param wg: "wg1", "wg2", "wg3 etc.
        :param chapter: Chapter01 etc
        :param infile: *.html_with_ids.html at present
        :param gv_output: output Graphviz file
        :param maxlev: maximum depth of sub/subsections (default 4)
        :param engine: layout engine (default fdp)
        """
        logger.info(f"reading from {infile=}")
        # parse infile to HTML object
        html = HtmlLib.parse_html(infile)
        assert html is not None
        # sections to be extracted using regexes
        WANTED_ID_RE = ".*acknowledgements|.*Executive|.*frequently\\-asked\\-questions|(n_)?\\d+(\\_\\d+)*"
        # sections to exclude
        UNWANTED_ID_RE = ".*siblings"
        # highest level div on chapter page
        top_div = html.xpath("/html/body//div[div[@class='h1-container']]")[0]
        assert top_div is not None
        # maximum depth to recurse
        top_id = wg
        if gv_output.exists():
            FileLib.delete_file(gv_output)
            # logger.info(f"file deleted {gv_output.absolute()=}")
        # assert not gv_output.exists(), f"deleted {gv_output.absolute()}"
        graph = graphviz.Digraph(top_id, filename=str(gv_output), engine=engine)
        graph.edge(wg, chapter)
        # iterate over children of top,
        for div in top_div.xpath("div[@class='h1-container']"):
            # get "id" attribute value
            id = cls.get_cleaned_id(div)
            # reject if id is not in selected list or in not-selected list
            if not re.match(WANTED_ID_RE, id) or re.match(UNWANTED_ID_RE, id):
                logger.info(f"skipped {id=}")
                continue
            # logger.info(f"add edge {chap}->{id}")
            graph.edge(chapter, id)
            cls.list_divs_with_ids(div, maxlevel=maxlev, graph=graph)
        if graph is not None:
            """NOTE the PDF viewer caches the displays and does not update
            this wasted a lot of time
            """
            graph.view()
        return graph


    @classmethod
    def list_divs_with_ids(cls, parent_div, maxlevel, graph=None):

        """
        recursively find descendants and create graph
        :param parent_div:
        :param maxlevel: decremented for each recursion
        :param graph: if not none, adds edges to graph
        """
        UNWANTED_ID_RE = ".*siblings"
        parent_id = cls.get_cleaned_id(parent_div)
        if maxlevel >= 0:
            # find children
            divs = parent_div.xpath("div")
            for div in divs:
                div_id = cls.get_cleaned_id(div)
                if div_id is None or re.match(UNWANTED_ID_RE, div_id):
                    logger.info(f"skipped {div_id=}")
                    continue
                # print(f"add node {div_id=}")
                # logger.info(f"add edge {graph} {parent_id}->{div_id}")
                if graph is not None:
                    graph.edge(parent_id, div_id)
                    # logger.info(f"edge: {graph} {parent_id}->{div_id}")
                cls.list_divs_with_ids(div, maxlevel-1, graph=graph)


    @classmethod
    def get_cleaned_id(self, div):
        """
        I think node ids should not look like (DP) numbers so replace "." and
        prefix with "n_"
        """
        id = div.get("id")
        id = id.replace("-", "_")
        id = id.replace(".", "_")
        id = id.replace("\"", "")
        if ":" in id:
            ids = id.split(":")
            id = ids[0]
        id = f"n_{id}"
        # logger.info(f"{id=}")

        return id







"""tests for file_lib"""
import glob
import os
import logging
import unittest

from pathlib import Path

from amilib.file_lib import FileLib
from amilib.util import Util

from test.resources import Resources
from test.test_all import AmiAnyTest

FILE_LIB = "file_lib"
PDF_LIB = "pdf_lib"
TEXT_LIB = "text_lib"
_SETUP = "_setup"
_TEARDOWN = "_teardown"
TEST = "test"

"""
a few tests on globbing, etc"""

logger = Util.get_logger(__name__)

class File0Test(AmiAnyTest):
    # nothing much here
    logger = logging.getLogger("test_file")
    TEST = "test"

    OPTIONS = [
        FILE_LIB,
        PDF_LIB,
        TEXT_LIB,
        _SETUP,
        _TEARDOWN,
    ]

    @classmethod
    def example_setup(cls, pyamix):
        """ setup test or examples
        these are general commands (i.e. not subcommands)
        :pyamix:
        """
        pyamix.run_commands([
            "--delete ${exam_temp}",
            "--copy ${examples_test.p} ${exam_temp} overwrite",
        ])

    @classmethod
    def example_teardown(cls, pyamix):
        """ clean example files

        """
        pyamix.run_command([
            "--delete", "${exam_temp}",
        ])

    @classmethod
    def run_arg_tests(cls, args):
        """This needs revision , maybe using Examples()"""
        cls.logger.warning(f"*****running tests : {args[TEST]}")
        if not args[TEST]:
            cls.logger.warning(f"No tests given: choose some/all of {TEST}")
            return
        if FILE_LIB in args[TEST]:
            cls.logger.warning("run test_file")
            cls.test_file.main()
        if PDF_LIB in args[TEST]:
            cls.logger.warning("run test_pdf")
            cls.test_pdf.test_read_pdf()
        if TEXT_LIB in args[TEST]:
            cls.logger.warning("run test_text NYI")


    @unittest.skip("if not windows")
    def test_glob_with_windows_FILE(self):
        """
        some instances of glob.glob with fully forward slashes generate backslash or double backslash on windows
        """
        input_dir = Path(Resources.TEST_RESOURCES_DIR, "ar6", "cleaned_content")
        assert input_dir.exists()
        glob_str = f"{str(input_dir)}/**/html_with_ids.html"
        files = FileLib.posix_glob(glob_str, recursive=True)
        assert len(files) > 70
        for file in files[:5]:
            logger.debug(f"file: {file}")
            assert not "\\" in file # backslash is not generated by UNIX/MacOS
            assert "/" in file

    def test_list_children(self):
        """
        test that children can be listed as dir/not_dir. Ignore hidden files as these
        are not very standard
        """
        top_dir = self.make_self_test_dir()
        if Path(top_dir).exists():
            FileLib.delete_directory_contents(top_dir, delete_directory=True)
        assert not top_dir.exists()
        top_dir.mkdir()
        assert top_dir.exists()
        dir1 = Path(top_dir, "dir1")
        dir1.mkdir()
        assert dir1.exists()
        file1 = Path(top_dir, "file1")
        file1.touch()
        assert file1.exists()
        files = [str(f) for f in FileLib.get_children(top_dir)]
        assert sorted(files) == ['/Users/pm286/workspace/amilib/temp/test_files/dir1',
                                 '/Users/pm286/workspace/amilib/temp/test_files/file1']
        files = [str(f) for f in FileLib.get_children(top_dir, dirx=True)]
        assert sorted(files) == ['/Users/pm286/workspace/amilib/temp/test_files/dir1']
        files = [str(f) for f in FileLib.get_children(top_dir, dirx=False)]
        assert sorted(files) == ['/Users/pm286/workspace/amilib/temp/test_files/file1']

    @classmethod
    @unittest.skip("not yet right")
    def test_expand_braces(cls):
        """
        no longer tests braces
        :return:
        """

        home = os.path.expanduser("~")
        file = __file__
        python_dir = os.path.abspath(file + "/../")
        resources_dir = os.path.join(python_dir, "resources")
        logger.debug("python", os.path.abspath(python_dir))
        physchem_dir = os.path.abspath(python_dir + "/../")
        open_diagram = os.path.abspath(physchem_dir + "/../")
        logger.debug(os.path.abspath(open_diagram))
        opend = os.path.join(open_diagram, "*")
        logger.debug("opend", opend)
        logger.debug(f"od0 {glob.glob(opend)}")
        open_diagram01 = os.path.join(physchem_dir, "**", "liion10", "**", "*.*ml")
        logger.debug(f"od01 {glob.glob(open_diagram01)}")

        # open_diagram02 = os.path.join(home, "projects", "*iagram", "{*ot*,*.md}")
        # logger.debug("od02", bg.braced_glob(open_diagram02))
        open_diagram11 = os.path.join(home, "projects", "*iagram", "*", "*", "*.xml")
        logger.debug("od", open_diagram11)
        logger.debug("b", glob.glob(open_diagram11))
        open_diagram12 = os.path.join(home, "projects", "*iagram", "**", "*.xml")
        logger.debug("od", open_diagram12)

    def test_get_input_strings(self):
        """
        reads lists of words with optional split
        """
        words = FileLib.get_input_strings("foo")
        assert words == ["foo"]
        words = FileLib.get_input_strings(["foo"])
        assert words == ["foo"]
        words = FileLib.get_input_strings(["foo", "bar"])
        assert words == ["foo", "bar"]
        words = FileLib.get_input_strings("foo bar")
        assert words == ["foo bar"]
        words = FileLib.get_input_strings("foo bar", split=True)
        assert words == ["foo", "bar"]

    def test_get_input_strings_from_file(self):
        """
        reads lists of words in file/s
        """
        words = FileLib.get_input_strings(Path(Resources.TEST_RESOURCES_DIR, "wordlists", "small_2.txt"))
        assert words == ["anthropogenic", "tropospheric"]
        words = FileLib.get_input_strings(Path(Resources.TEST_RESOURCES_DIR, "wordlists", "small_10.txt"))
        assert words == [
            'anthropogenic', 'physical-biogeochemical', 'peat drainage', 'tropospheric', 'permafrost',
            'centennial', 'aerosols', 'sequestration', 'albedo', 'exacerbating'
        ]
        words = FileLib.get_input_strings(Path(Resources.TEST_RESOURCES_DIR, "wordlists", "small_10.txt"), split=True)
        # assert words == [
        #     'anthropogenic', 'physical-biogeochemical', 'peat', 'drainage', 'tropospheric', 'permafrost',
        #     'centennial', 'aerosols', 'sequestration', 'albedo', 'exacerbating'
        # ]
        assert words ==['anthropogenic',
 'physical-biogeochemical',
 'peat',
 'drainage',
 'tropospheric',
 'permafrost',
 'centennial',
 'aerosols',
 'sequestration',
 'albedo',
 'exacerbating']
        # list of files
        words = FileLib.get_input_strings([
            Path(Resources.TEST_RESOURCES_DIR, "wordlists", "small_5.txt"),
            Path(Resources.TEST_RESOURCES_DIR, "wordlists", "small_10.txt")])

        # assert words == ['anthropogenic', 'physical-biogeochemical', 'peat drainage', 'tropospheric',
        #     'permafrost', 'centennial', 'aerosols', 'sequestration', 'albedo', 'exacerbating']
        assert words == ['anthropogenic',
 'tropospheric',
 'permafrost',
 'sequestration',
 'albedo',
 'anthropogenic',
 'physical-biogeochemical',
 'peat drainage',
 'tropospheric',
 'permafrost',
 'centennial',
 'aerosols',
 'sequestration',
 'albedo',
 'exacerbating']

        # list of files
        words = FileLib.get_input_strings([
            Path(Resources.TEST_RESOURCES_DIR, "wordlists", "small_5.txt"),
            Path(Resources.TEST_RESOURCES_DIR, "wordlists", "small_10.txt")], split=True)

        assert words == ['anthropogenic',
 'tropospheric',
 'permafrost',
 'sequestration',
 'albedo',
 'anthropogenic',
 'physical-biogeochemical',
 'peat',
 'drainage',
 'tropospheric',
 'permafrost',
 'centennial',
 'aerosols',
 'sequestration',
 'albedo',
 'exacerbating']


    def test_relative_pathname(self):
        """
        convert absolute pathname to relative to another path
        some of this is in Python 3.12 but we'll support earlier versions
        """
        abspath = Path("/a/b/c/d/e.txt")
        refpath = Path("/a/b/c/")
        diffpath = FileLib.get_reletive_path(abspath, refpath)
        assert str(diffpath) == "d/e.txt"

        diffpath = FileLib.get_reletive_path(refpath, abspath)
        assert str(diffpath) == "../.."

        # dont' use Python 3.12 walk_up
        refpath = Path("/a/b/c/x/y.txt")
        try:
            diffpath = FileLib.get_reletive_path(abspath, refpath, walk_up=False)
        except ValueError as e:
            assert str(e) == "'/a/b/c/d/e.txt' is not in the subpath of '/a/b/c/x/y.txt'"

        # use Python 3.12 walk_up
        refpath = Path("/a/b/c/x/y/z.txt")

        diffpath = FileLib.get_reletive_path(abspath, refpath, walk_up=True)
        assert str(diffpath) == ("../../../d/e.txt")
        diffpath = FileLib.get_reletive_path(refpath, abspath, walk_up=True)
        assert str(diffpath) == ("../../x/y/z.txt")

    def test_force_write_operations(self):
        """Test basic file writing operations"""
        test_file = self.make_self_test_dir() / "test.txt"
        
        # Test basic write
        FileLib.force_write(test_file, "hello")
        self.assertTrue(test_file.exists())
        self.assertEqual(test_file.read_text(), "hello")
        
        # Test overwrite=False
        FileLib.force_write(test_file, "world", overwrite=False)
        self.assertEqual(test_file.read_text(), "hello")  # Should not change
        
        # Test nested path creation
        nested_file = self.test_dir / "a/b/c/test.txt"
        FileLib.force_write(nested_file, "nested")
        self.assertTrue(nested_file.exists())
        self.assertEqual(nested_file.read_text(), "nested")

    def test_copy_operations(self):
        """Test file and directory copying"""
        # Setup test files

        # source_dir = self.test_dir / "source" PMR - there is no self.test_dir
        source_dir = self.make_self_test_dir() / "source"
        source_dir.mkdir(exist_ok=True)
        assert source_dir.is_dir()
        test_txt = source_dir / "test.txt"
        test_txt = (test_txt)
        logger.debug(f"abs path {test_txt.absolute()}")
        test_txt.write_text("hello")
        assert test_txt.is_file() and not test_txt.is_dir()
        self.assertEqual(test_txt.read_text(), "hello")
        (source_dir / "subdir").mkdir(exist_ok=True)
        (source_dir / "subdir/test2.txt").write_text("world")
        
        # Test file copy
        dest_file = Path(self.test_dir, "dest.txt")
        FileLib.copy_file_or_directory(
            dest_file,
            test_txt,
            overwrite=True
        )
        self.assertTrue(dest_file.exists())
        self.assertEqual(dest_file.read_text(), "hello")
        
        # Test directory copy
        dest_dir = self.test_dir / "dest"
        FileLib.copy_file_or_directory(
            dest_dir,
            source_dir,
            overwrite=True
        )
        self.assertTrue((dest_dir / "test.txt").exists())
        self.assertTrue((dest_dir / "subdir/test2.txt").exists())

    def test_input_string_edge_cases(self):
        """Test edge cases for input string handling"""
        # Test empty inputs
        self.assertEqual(FileLib.get_input_strings(None), [])
        self.assertEqual(FileLib.get_input_strings([]), [])
        
        # Test mixed input types
        self.make_self_test_dir()
        test_file = self.test_dir / "words.txt"
        test_file.write_text("one\ntwo")
        
        result = FileLib.get_input_strings([
            "hello",
            test_file,
            "world"
        ])
        self.assertEqual(
            result,
            ["hello", "one", "two", "world"]
        )

    def make_self_test_dir(self):
        try:
            self.test_dir = self.__getattribute__("test_dir")
        except AttributeError as ae:
            self.test_dir = None
        if self.test_dir is None:
            self.test_dir = Path(Resources.TEMP_DIR, "test_files")
        self.test_dir.mkdir(exist_ok=True)
        return self.test_dir


def main():
    File0Test.test_file_simple()
    File0Test.test_templates()



import csv
import logging
import re
import unittest
from collections import Counter
from pathlib import Path

import lxml
import lxml.etree as ET
from amilib.file_lib import FileLib
from amilib.html_generator import HtmlGenerator
#
from amilib.html_marker import SpanMarker, HtmlPipeline
from amilib.util import Util
from amilib.ami_html import HtmlLib

# from climate.amix import AMIClimate, REPO_DIR
# from climate.un import DECISION_SESS_RE, MARKUP_DICT, INLINE_DICT, UNFCCC, UNFCCCArgs
#
from test.resources import Resources
from test.test_all import AmiAnyTest

"""Mainly to test PDF conversion (which used to work)"""
#
# required for UNFCCC which was developed in amiclimat
REPO_DIR = Path(Resources.TEST_RESOURCES_DIR).parent

UNFCCC_DIR = Path(Resources.TEST_RESOURCES_DIR, "unfccc")
UNFCCC_TEMP_DIR = Path(Resources.TEMP_DIR, "unfccc")
UNFCCC_TEMP_DOC_DIR = Path(UNFCCC_TEMP_DIR, "unfcccdocuments1")

ROMAN = "I|II|III|IIII|IV|V|VI|VII|VIII|IX|X|XI|XII|XIII|XIV|XV|XVI*"
L_ROMAN = "i|ii|iii|iv|v|vi|vii|viii|ix|x|xi|xii|xiii|xiv|xv|xvi|xvii|xviii|xix|xx"
INT = "\\d+"  # integer of any length
DIGIT = "\\d"  # single digit
DOT = f"\\."  # dot
MINUS = "-"
FLOAT = f"{MINUS}?{INT}({DOT}{INT})?"
SP = "\\s"  # single space
WS = "\\s+"  # spaces
ANY = ".*"
SL = "/"  # slash
LP = "\\("  # left parenthesis
RP = "\\)"  # right parenthesis
LC = "[a-z]"  # single uppercase
UC = "[A-Z]"  # single uppercase
#
DECISION_SESS_RE = re.compile(
    f"(?P<front>{ANY}\\D)(?P<dec_no>{INT})/(?P<body>{ANY}){DOT}(?P<sess_no>{INT}){DOT}?(?P<end>{ANY})")

RESERVED_WORDS = {
    'Recalling',
    'Also recalling',
    'Further recalling',
    'Recognizing',
    'Cognizant',
    'Annex',
    'Abbreviations and acronyms',
    'Noting',
    'Acknowledging',
}

RESERVED_WORDS1 = "(Also|[Ff]urther )?([Rr]ecalling|[Rr]ecogniz(es|ing)|Welcomes|[Cc]ognizant|[Nn]ot(ing|es)" \
                  "|Invit(es|ing)|Acknowledging|[Ex]pressing appreciation]|Recalls|Stresses|Urges|Requests|Expresses alarm)"

CPTYPE = "CP|CMA|CMP"
SUBPARA = f"({LP}?P<subpara>{LC}){RP}"
SUBSUBPARA = f"({LP}?P<subsubpara>{L_ROMAN}){RP}"
PARENT_DIR = "unfccc/unfcccdocuments1"  # probably temporary
TARGET_DIR = "../../../../../temp/unfccc/unfcccdocuments1/"

REPO_TOP = "https://raw.githubusercontent.com/petermr/pyamihtml/main"
TEST_REPO = f"{REPO_TOP}/test/resources/unfccc/unfcccdocuments1"
TEMP_REPO = f"{REPO_TOP}/temp/unfccc/unfcccdocuments1"

# markup against terms in spans
TARGET_STEM = "marked"  # was "split"

INLINE_DICT = {
    "decision": {
        "example": ["decision 1/CMA.2", "noting decision 1/CMA.2, paragraph 10 and ", ],
        "regex":
        # f"decision{WS}(?P<decision>{INT})/(?P<type>{CPTYPE}){DOT}(?P<session>{INT})",
        # f"decision{WS}(?P<decision>{INT})/(?P<type>{CPTYPE}){DOT}(?P<session>{INT})(,{WS}paragraph(?P<paragraph>{WS}{INT}))?",
            f"(?P<decision>{INT})/(?P<type>{CPTYPE}){DOT}(?P<session>{INT})",
        "href": "FOO_BAR",
        "split_span": True,
        "idgen": "NYI",
        "_parent_dir": f"{TARGET_DIR}",
        "span_range": [0, 99],

        # "href_template": f"{PARENT_DIR}/{{type}}_{{session}}/Decision_{{decision}}_{{type}}_{{session}}",
        # "href_template": f"../../{{type}}_{{session}}/Decision_{{decision}}_{{type}}_{{session}}",
        "href_template": f"{TARGET_DIR}/{{type}}_{{session}}/Decision_{{decision}}_{{type}}_{{session}}/{TARGET_STEM}.html",
    },
    "paragraph": {
        "example": [
            "paragraph 32 above",
            "paragraph 23 below",
            "paragraph 9 of decision 19/CMA.3",
            "paragraph 77(d)(iii)",
            "paragraph 37 of chapter VII of the annex",
        ],
        "regex": [f"paragraph (?P<paragraph>{INT} (above|below))",
                  f"paragraph (?P<paragraph>{INT}{LP}{LC}{RP}{LP}{L_ROMAN}{RP})"
                  ],
    },
    "exhort": {
        "regex": f"{RESERVED_WORDS1}",
        "href": "None",
    },
    "article": {
        "example": ["Article 4, paragraph 19, of the (Paris Agreement)",
                    "tenth preambular paragraph of the Paris Agreement",
                    "Article 6, paragraph 3"],
        "regex": f"Article (?P<article>{INT}), paragraph (?P<paragraph>{INT}), (of the (?P<agreement>Paris Agreement))?",
    },
    "trust_fund": {
        "regex": "Trust Fund for Supplementary Activities",
        "href_template": "https://unfccc.int/documents/472648",
    },
    "adaptation_fund": {
        "regex": "([Tt]he )?Adaptation Fund",
        "href_template": "https://unfccc.int/Adaptation-Fund",
    },
    "paris": {
        "regex": "([Tt]he )?Paris Agreement",
        "href_template": "https://unfccc.int/process-and-meetings/the-paris-agreement",
    },
    "cop": {
        "regex": "([Tt]he )?Conference of the Parties",
        "href_template": "https://unfccc.int/process/bodies/supreme-bodies/conference-of-the-parties-cop",
    },
    "sbi": {
        "regex": "([Tt]he )?Subsidiary Body for Implementation",
        "acronym": "SBI",
        "wiki": "https://en.wikipedia.org/wiki/Subsidiary_Body_for_Implementation",
        "href": "https://unfccc.int/process/bodies/subsidiary-bodies/sbi"
    },
    # data
    "temperature": {
        "example": "1.5 °C",
        "regex": f"{FLOAT}{WS}°C",
        "class": "temperature",
    },
    # date
    "date": {
        "example": "2019",
        "regex": f"20\\d\\d",
        "class": "date",
    }
}
# section dict
MARKUP_DICT = {
    "Decision": {
        "level": 0,
        "parent": [],
        "example": ["Decision 1/CMA.1", "Decision 1/CMA.3"],
        "regex": f"Decision (?P<Decision>{INT})/(?P<type>{CPTYPE})\\.(?P<session>{INT})",
        "components": ["", ("Decision", f"{INT}"), "/", ("type", {CPTYPE}), f"{DOT}", ("session", f"{INT}"), ""],
        "names": ["roman", "title"],
        "class": "Decision",
        "span_range": [0, 1],
        "template": "Decision_{Decision}_{type}_{session}",
    },
    "Resolution": {
        "level": 0,
        "parent": [],
        "example": ["Resolution 1/CMA.1", "Resolution 1/CMA.3"],
        "regex": f"Resolution (?P<Resolution>{INT})/(?P<type>{CPTYPE})\\.(?P<session>{INT})",
        "components": ["", ("Resolution", f"{INT}"), "/", ("type", {CPTYPE}), f"{DOT}", ("session", f"{INT}"), ""],
        "names": ["roman", "title"],
        "class": "Resolution",
        "span_range": [0, 1],
        "template": "Resolution{Resolution}_{type}_{session}",
    },
    "chapter": {
        "level": 1,
        "parent": ["Decision"],
        "example": ["VIII.Collaboration", "I.Science and urgency"],
        "regex": f"(?P<dummy>)(?P<roman>{ROMAN}){DOT}\\s*(?P<title>{UC}.*)",
        "components": [("dummy", ""), ("roman", f"{ROMAN}"), f"{DOT}{WS}", ("title", f"{UC}{ANY}")],
        "names": ["roman", "title"],
        "class": "chapter",
        "span_range": [0, 1],
        "template": "chapter_{roman}",
    },
    "subchapter": {
        "level": "C",
        "parent": ["chapter"],
        "example": ["B.Annual information"],
        "regex": f"(?P<capital>{UC}){DOT}",
        "names": ["subchapter"],
        "class": "subchapter",
        "span_range": [0, 1],
        "template": "subchapter_{capital}",
    },

    "para": {
        "level": 2,
        "parent": ["chapter", "subchapter"],
        "example": ["26. "],
        "regex": f"(?P<para>{INT}){DOT}{SP}*",
        "names": ["para"],
        "class": "para",
        "parent": "preceeding::div[@class='roman'][1]",
        "idgen": {
            "parent": "Decision",
            "separator": ["_", "__"],
        },
        "span_range": [0, 1],
        "template": "para_{para}",
    },
    "subpara": {
        "level": 3,
        "parent": ["para"],
        "example": ["(a)Common time frames"],
        "regex": f"{LP}(?P<subpara>{LC}){RP}",
        "names": ["subpara"],
        "class": "subpara",
        "span_range": [0, 1],
        "template": "subpara_{subpara}",

    },
    "subsubpara": {
        "level": 4,
        "parent": ["subpara"],
        "example": ["(i)Methods for establishing"],
        "regex": f"{LP}(?P<subsubpara>{L_ROMAN}){RP}",
        "names": ["subsubpara"],
        "class": "subsubpara",
        "span_range": [0, 1],
    },

}
DECISION_SESS_RE = ""
#
MAXPDF = 3
#
# OMIT_LONG = True  # omit long tests
#
# #
TEST_DIR = Path(REPO_DIR, "test")
TEMP_DIR = Path(REPO_DIR, "temp")
#
IPCC_TOP = Path(TEST_DIR, "resources", "ipcc", "cleaned_content")
# assert IPCC_TOP.exists(), f"{IPCC_TOP} should exist"
#
QUERIES_DIR = Path(TEMP_DIR, "queries")
# assert QUERIES_DIR.exists(), f"{QUERIES_DIR} should exist"
#
IPCC_DICT = {
    "_IPCC_REPORTS": IPCC_TOP,
    "_IPCC_QUERIES": QUERIES_DIR,
}
#
CLEANED_CONTENT = 'cleaned_content'
SYR = 'syr'
SYR_LR = 'longer-report'
IPCC_DIR = 'ipcc'
#

OUT_DIR_TOP = Path(FileLib.get_home(), "workspace")

# input
IPCC_URL = "https://www.ipcc.ch/"
AR6_URL = IPCC_URL + "report/ar6/"
SYR_URL = AR6_URL + "syr/"
WG1_URL = AR6_URL + "wg1/"
WG2_URL = AR6_URL + "wg2/"
WG3_URL = AR6_URL + "wg3/"

# SC_TEST_DIR = Path(OUT_DIR_TOP, "ipcc", "ar6", "test")

# logger = logging.getLogger(__file__)
# logger.setLevel(logging.WARNING)
logger = Util.get_logger(__name__)
logger.setLevel(logging.INFO)


class TestUNFCCC(AmiAnyTest):
    """Tests high level operations relating to UN content (currently SpanMarker and UN/IPCC)
    """

    STYLES = [
        #  <style classref="div">div {border: red solid 0.5px;}</style>
        "span.temperature {border: purple solid 0.5px;}",
        ".chapter {border: blue solid 0.8px; font-weight: bold; background: red;}",
        ".subchapter {background: pink;}",
        ".para {border: blue dotted 0.6px; margin: 0.3px;}",
        ".subpara {border: blue solid 0.4px; margin: 0.2px; background: #eeeeee; opacity: 0.7}",
        ".subsubpara {border: blue dashed 0.2px; margin: 2px; background: #dddddd; opacity: 0.3}",
        "a[href] {background: #ffeecc;}",
        "* {font-size: 7; font-family: helvetica;}",
    ]

    UNFCCC_DICT = {
        # "1" : {
        #
        # },
        # "*" : {
        "name": "UNFCCC reports",
        "footer_height": 50,  # from lowest href underlines
        "header_height": 70,  # from 68.44
        "header_bottom_line_xrange": [20, 700],
        "footnote_top_line_xrange": [50, 300],
        "box_as_line_height": 1
    }

    @classmethod
    def create_initial_directories(cls, in_sub_dir, in_file, top_out_dir, out_stem=None, out_suffix="html"):
        """creates initial directory structure from PDF files
        in:
        test/resources/unfccc/unfcccdocuments1/CMA_3/1_4_CMA_3.pdf
        ............................top_in_dir| (implicit filename.parent.parent)
        ....................................._in_file............|
        ..................................in_sub_dir|
                                                    |.........| = in_subdir_stem
        out:
        temp/unfccc/unfcccdocuments1/CMA_3/1_4_CMA_3/raw.html
        .................top_outdir|
        ........................out_subdir|
        ...............................out_subsubdir|
                                                    |...|= out_stem
                                                         |....| = out_suffix

        Create outsubdir with same stem as in in_subdir
        create out_subsubdir from in_file stem
        create any directories with mkdir()
        :param in_sub_dir: subdirectory of corpus (session)
        :param in_file: file has implict subsub in stem
        :param top_out_dir: top of output directory (analogous to top_in_dir)
        :param out_stem: no default
        :param out_suffix: defaults to "html"
        :return: out_subsubdir, outfile (None if out_stem not given)

        thus
        """
        in_subdir_stem = Path(in_sub_dir).stem
        out_subdir_stem = in_subdir_stem
        out_subdir = Path(top_out_dir, out_subdir_stem)
        out_subdir.mkdir(parents=True, exist_ok=True)
        out_subsubdir = Path(out_subdir, in_file.stem)
        out_subsubdir.mkdir(parents=True, exist_ok=True)
        outfile = Path(out_subsubdir, out_stem + "." + out_suffix) if out_stem else None
        return out_subsubdir, outfile

    @classmethod
    def run_pipeline_on_unfccc_session(
            cls,
            in_dir,
            session_dir,
            in_sub_dir=None,
            top_out_dir=None,
            file_splitter=None,
            targets=None,
            directory_maker=None,
            markup_dict=None,
            inline_dict=None,
            param_dict=None,
            styles=None
    ):
        """
        directory structure is messy
        """

        session = Path(session_dir).stem
        if in_sub_dir is None:
            in_sub_dir = Path(in_dir, session)
        pdf_list = FileLib.posix_glob(str(in_sub_dir) + "/*.pdf")
        print(f"pdfs in session {session} => {pdf_list}")
        if not pdf_list:
            print(f"****no PDFs in {in_sub_dir}")
        subsession_list = [Path(pdf).stem for pdf in pdf_list]
        print(f"subsession_list {subsession_list}")
        if not top_out_dir:
            print(f"must give top_out_dir")
            return
        out_sub_dir = Path(top_out_dir, session)
        skip_assert = True
        if not file_splitter:
            file_splitter = "span[@class='Decision']"  # TODO move to dictionary
        if not targets:
            targets = ["decision", "paris", "wmo", "temperature"]
        if not directory_maker:
            directory_maker = TestUNFCCC
        if not markup_dict:
            markup_dict = MARKUP_DICT
        if not inline_dict:
            inline_dict = INLINE_DICT
        if not param_dict:
            param_dict = TestUNFCCC.UNFCCC_DICT
        if not styles:
            styles = TestUNFCCC.STYLES
        for subsession in subsession_list:
            HtmlPipeline.stateless_pipeline(

                file_splitter=file_splitter, in_dir=in_dir, in_sub_dir=in_sub_dir, instem=subsession,
                out_sub_dir=out_sub_dir,
                top_out_dir=top_out_dir,
                page_json_dir=Path(top_out_dir, "json"),
                directory_maker=directory_maker,
                markup_dict=markup_dict,
                inline_dict=inline_dict,
                param_dict=param_dict,
                targets=targets,
                styles=styles,
                force_make_pdf=True)

    @classmethod
    def create_decision_table(cls, in_dir, outcsv, outcsv_wt=None):
        """create table of links, rather ad hoc"""
        decision_files = TestUNFCCC.extract_decision_files_posix(in_dir)
        weight_dict = Counter()
        typex = "type"
        for decision_file in decision_files:
            decision_path = Path(decision_file)
            a_els = TestUNFCCC.extract_hyperlinks_to_decisions(decision_file)
            source_id = str(decision_path.parent.stem).replace("ecision", "")
            for a_elem in a_els:
                text = a_elem.text
                splits = text.split(",")
                # this should use idgen
                target_id = splits[0].replace("d", "D").replace(" ", "_").replace("/", "_").replace(".", "_") \
                    .replace("ecision", "")
                para = splits[1] if len(splits) == 2 else ""
                edge = (source_id, target_id, para)
                weight_dict[edge] += 1
        with open(outcsv, "w") as fw:
            FileLib.force_mkdir(outcsv.parent)
            csvwriter = csv.writer(fw)
            csvwriter.writerow(["source", "link_type", "target", "para", "weight"])
            for (edge, wt) in weight_dict.items():
                csvwriter.writerow([edge[0], typex, edge[1], edge[2], wt])
        print(f"wrote {outcsv}")

    @classmethod
    def extract_hyperlinks_to_decisions(self, marked_file):
        """Currently hyperlinks are
        file:///Users/pm286/workspace/pyamihtml_top/temp/unfccc/unfcccdocuments1//CP_21/Decision_1_CP_21/marked.html
        <a href="../../../../../temp/unfccc/unfcccdocuments1//CMA_3/Decision_1_CMA_3/marked.html">1/CMA.3</a>
        """

        html_elem = lxml.etree.parse(str(marked_file))
        a_elems = html_elem.xpath(".//a[@href[contains(.,'ecision')]]")
        return a_elems

    @classmethod
    def extract_decision_files_posix(cls, in_dir, stem="marked"):
        """extracts all files with "Decision" in file name
        :param in_dir: top directory of corpus (immediate children are session directories e.g. CMP_3
        :param stem: file stem, e.g. 'split', 'marked'"""
        files = FileLib.posix_glob(str(in_dir) + f"/*/Decision*/{stem}.html")
        return files


    # ========================== tests ==================

    # @unittest.skip("not the current approach. TODO add make to spanmarker pipeline")
    def test_read_unfccc_everything_MAINSTREAM(self):
        """"""
        """
        * reads unfccc reports in PDF,
        * transdlates to HTML,
        * adds semantic indexing for paragraphs
        * writes markedup html
        * extracts targets from running text (NYI)
        * builds csv table (NYI)
        which can be fed into pyvis to create a knowledge graph
        (writes outout to wrong dir)
        MAINSTREAM!
        """
        """
        Doesn't out
        ut anything
        """
        input_dir = Path(UNFCCC_DIR, "unfcccdocuments1")
        pdf_list = FileLib.posix_glob(f"{input_dir}/*C*/*.pdf")[:MAXPDF]  # select CMA/CMP/CP
        outcsv = "links.csv"
        outdir = Path(Resources.TEMP_DIR, "unfcccOUT")
        outhtmldir = str(Path(outdir, "newhtml"))
        markup_dict = MARKUP_DICT

        span_marker = SpanMarker(regex=DECISION_SESS_RE)
        span_marker.run_pipeline(input_dir=input_dir,
                                 outcsv=outcsv,
                                 outdir=outdir,
                                 pdf_list=pdf_list,
                                 markup_dict=markup_dict,
                                 outhtml=outhtmldir,
                                 debug=True
                                 )

    # @unittest.skip("probably redundant")
    def test_convert_pdfs_to_raw_html_IMPORTANT_STEP_1(self):
        """
        FIRST OPERATION
        tests reading the whole PDFs
        creates HTML elements
        OUTPUT - RAW HTML
        (raw HTML contains raw styles (e.g. .s1 ... .s78) in head/style)
        <style>.s15 {font-family: Times New Roman; font-size: 9.96;}</style>
        NOT normalized so we get
        <style>.s16 {font-family: Times New Roman; font-size: 9.96;}</style>

        steps:

        html_elem = HtmlGenerator.convert_to_html("foo", pdf)

        """

        input_dir = Path(UNFCCC_DIR, "unfcccdocuments1")
        assert input_dir.exists()
        pdfs = FileLib.posix_glob(f"{input_dir}/*C*/*.pdf")[:MAXPDF]
        assert MAXPDF >= len(pdfs) > 0
        for pdf in pdfs:
            print(f"parsing {pdf}")
            html_elem = HtmlGenerator.read_pdf_convert_to_html(pdf)

            # assertions
            assert html_elem is not None
            print (f"html {ET.tostring(html_elem)[:400]}")
            # does element contain styles?
            head = HtmlLib.get_head(html_elem)
            assert head is not None
            """
<head>
    <style>div {border : red solid 0.5px}</style>
    ...
</head>
"""
            styles = head.xpath("style")
            assert len(styles) > 5
            # are there divs?
            """
    <div left="113.42" right="123.75" top="451.04">
        <span x0="113.42" y0="451.04" x1="123.75" style="x0: 113.42; x1: 118.4; y0: 451.04; y1: 461.0; width: 4.98;" class="s34">1. </span>
        <span x0="141.74" y0="451.04" x1="184.67" style="x0: 141.74; x1: 150.04; y0: 451.04; y1: 461.0; width: 8.3;" class="s35">Welcomes </span>
        <span x0="184.7" y0="451.04" x1="451.15" style="x0: 184.7; x1: 187.47; y0: 451.04; y1: 461.0; width: 2.77;" class="s36">the entry into force of the Paris Agreement on 4 November 2016;  </span>
    </div>"""
            divs = HtmlLib.get_body(html_elem).xpath("div")
            assert len(divs) > 5
            # do they have spans with styles?
            spans = HtmlLib.get_body(html_elem).xpath("div/span[@class]")
            assert len(spans) > 20

            # outdir, outfile = SpanMarker.create_dir_and_file(pdf, stem="raw", suffix="html")
            outfile = f"{pdf}.raw.html"
            HtmlLib.write_html_file(html_elem, outfile=outfile, debug=True)

            assert Path(outfile).exists()

    def test_find_unfccc_decisions_INLINE_markup_regex_single_document_IMPORTANT(self):
        """
        INLINE marking
        looks for strings such as decision 20/CMA.3 using regex
        single

        takes simple HTML element and marks it with the in-span "decision"
        div
            span
        and splits the span with a regex, annotating the results
        adds classes
        tackles most of functionality



        """
        """
        Does inline markup
        """

        html_infile = self._normalized_test_file()
        targets = [
            "decision",
            "paris",
            # "adaptation_fund"
        ]

        span_marker = SpanMarker()
        span_marker.split_spans_in_html(html_infile=html_infile, targets=targets, markup_dict=INLINE_DICT, debug=True)

    def _normalized_test_file(self):
        UNFCCC_TEMP_DOC_DIR = Path(Resources.TEST_RESOURCES_DIR, "unfccc", "unfcccdocuments1")
        assert UNFCCC_TEMP_DOC_DIR.exists()
        input_dir = Path(UNFCCC_TEMP_DOC_DIR, "CMA_3")
        html_infile = Path(input_dir, "1_4_CMA_3", "normalized.html")  # not marked
        return html_infile

    def test_inline_dict_IMPORTANT(self):
        """
        This matches keywords but doesn't markup file .
        DOESNT do hyperlinks
        """
        html_infile = self._normalized_test_file()
        input_dir = html_infile.parent
        html_outdir = Path(Resources.TEMP_DIR, "unfccc", "html")
        span_marker = SpanMarker(markup_dict=INLINE_DICT)
        outfile = Path(html_outdir, "1_CMA_3", "normalized.marked.html")
        Util.delete_file_and_check(outfile)
        html_elem = ET.parse(str(html_infile))
        span_marker.markup_html_element_with_markup_dict(
            html_elem,
            input_dir=input_dir,
            html_outdir=html_outdir,
            dict_name="dummy_dict",
            html_out=outfile,
            debug=True
        )
        assert outfile.exists()
        html_out_elem = HtmlLib.parse_html(outfile)
        assert html_out_elem is not None, f"html_out_elem from {outfile} should not be None"
        ahrefs = html_out_elem.xpath(".//a/@href")
        print(f"hrefs: {len(ahrefs)}")

    def test_find_ids_markup_dict_single_document_IMPORTANT_2023_01_01(self):

        """TODO generate ids of section tags"""
        """
        looks for strings , especially with IDs ,
        single

        takes simple HTML element and marks it with the MARKUP_DICT
        div
            span
        and splits the span with a regex, annotating the results
        adds classes
        tackles most of functionality

        """
        """output_id of form DEC_1_CMA_3__VII__78__b"""
        """output_id of form RES_1_CMA_3__VII__78__b__iv"""
        """INPUT is HTML"""
        """WORKS - outputs marked up sections in files"""
        # regex = "[Dd]ecisions? \s*\d+/(CMA|CP)\.\d+"  # searches for strings of form fo, foo, for etc
        dict_name = "sections"

        html_infile = self._normalized_test_file()
        input_dir = html_infile.parent

        html_outdir = Path(Resources.TEMP_DIR, "unfccc", "html")
        outfile = Path(html_outdir, "1_4_CMA_3", f"split.{dict_name}.html")
        markup_dict = MARKUP_DICT
        html_elem = SpanMarker.markup_file_with_markup_dict(
            input_dir, html_infile, html_outdir=html_outdir, dict_name=dict_name, outfile=outfile,
            markup_dict=markup_dict, debug=True)
        assert outfile.exists()
        assert len(HtmlLib.get_body(html_elem).xpath("div")) > 0

    @unittest.skip("obsolete approach to splitting files. TODO needs mending")
    def test_split_into_files_at_id_single_IMPORTANT(self):

        dict_name = "sections"
        input_dir = Path(UNFCCC_TEMP_DOC_DIR, "CMA_3")
        infile = Path(input_dir, "1_CMA_3_section", f"normalized.{dict_name}.html")

        splitter = "./span[@class='Decision']"
        output_dir = input_dir
        SpanMarker.presplit_by_regex_into_sections(infile, output_dir, splitter=splitter, debug=True)

    @unittest.skip("until we fix the previous")
    def test_split_into_files_at_id_multiple_IMPORTANT(self):
        """Splits files at Decisions for all sessions"""
        """requires previous test to have been run"""

        dict_name = "sections"
        splitter = "./span[@class='Decision']/a/@href"
        MAXFILE = 3

        top_dir = Path(UNFCCC_DIR, "unfcccdocuments1")
        files = FileLib.posix_glob(str(top_dir) + "/*/*/normalized.html")
        assert len(files) > 0
        for infile in files[:MAXFILE]:
            print(f"infile {infile} ")
            session_dir = Path(infile).parent.parent
            print(f"session {session_dir}")
            output_dir = session_dir
            SpanMarker.presplit_by_regex_into_sections(infile, output_dir, splitter=splitter)

    @unittest.skip("Not finshed")
    def test_make_nested_divs(self):
        """IMPORTANT not finished"""
        """initial div files are 'flat' - all divs are siblings, Use parents in markup_dict to assemble
        """
        input_dir = Path(UNFCCC_DIR, "unfcccdocuments1", "CMA_3")
        infile = Path(input_dir, "1_4_CMA_3_section", f"normalized.sections.html")
        assert str(infile).endswith(
            "test/resources/unfccc/unfcccdocuments1/CMA_3/1_4_CMA_3_section/normalized.sections.html"),\
            f"should have a file .../test/resources/unfccc/unfcccdocuments1/CMA_3/1_4_CMA_3_section/normalized.sections.html"
        span_marker = SpanMarker(markup_dict=MARKUP_DICT)
        span_marker.parse_html(infile)
        span_marker.move_implicit_children_to_parents(span_marker.inhtml)
        outfile = str(infile).replace("sections", "nested")
        HtmlLib.write_html_file(span_marker.inhtml, outfile, debug=True)


    @unittest.skip("needs mending")
    def test_pipeline(self):
        """
        sequential operations
        input set of PDFs , -> raw.html -> id.html
        """

        print("lacking markup_dict")
        return
        # input dir of raw (unsplit PDFs) . Either single decisions or concatenated ones
        indir = Path(UNFCCC_DIR, "unfcccdocuments1")

        subdirs = FileLib.posix_glob(str(indir) + "/" + "C*" + "/")  # docs of form <UNFCCC_DIR>/C*/

        assert len(subdirs) == 12  # CMA_1 ... CMA_2... CP_27
        pdf_list = FileLib.posix_glob(subdirs[0] + "/" + "*.pdf")  # only CMA_1 to start with
        assert len(pdf_list) == 4
        # contains 4 PDFs as beloe
        skip = True
        if not skip:
            # TODO use symbolic top directory
            unittest.TestCase().assertListEqual(sorted(pdf_list), [
                '/Users/pm286/workspace/pyamihtml_top/test/resources/unfccc/unfcccdocuments1/CMA_1/13_20_CMA_1.pdf',
                '/Users/pm286/workspace/pyamihtml_top/test/resources/unfccc/unfcccdocuments1/CMA_1/1_CMA_1.pdf',
                '/Users/pm286/workspace/pyamihtml_top/test/resources/unfccc/unfcccdocuments1/CMA_1/2_CMA_1.pdf',
                '/Users/pm286/workspace/pyamihtml_top/test/resources/unfccc/unfcccdocuments1/CMA_1/3_12_CMA_1.pdf'
            ])

        # class for processing SpanMarker documents
        span_marker = SpanMarker(regex=DECISION_SESS_RE)

        span_marker.indir = '/Users/pm286/workspace/amiclimate/test/resources/unfccc/unfcccdocuments1'  # inout dir
        span_marker.outdir = Path(Resources.TEMP_DIR, "unfcccOUT")
        print(f"output to dir: {span_marker.outdir}")
        span_marker.outfile = "links.csv"  # probably in wrong place
        # convert to raw HTML
        span_marker.read_and_process_pdfs(pdf_list)
        span_marker.write_links("links.csv")  # currently no-op
        span_marker.analyse_after_match_NOOP()


    def test_explicit_conversion_pipeline_IMPORTANT_DEFINITIVE(self):
        """reads PDF and sequentially applies transformation to generate increasingly semantic HTML
        define output structure
        1 ) read a PDF with several concatenated Decisions and convert to raw html incluing paragraph-like divs => raw.html
           a) clip / extract headers and footers
           b) footnotes
        2 ) extract styles into head (one style per div)  combined
        3 ) normalize styles syntactically => normalized.html (syntactic styles)
        4 ) tag divs by style and content
        5 ) split major sections into separate HTML files (CMA1_4 -> CMA1, CMA2 ...)
        6 ) obsolete
        7 ) assemble hierarchical documents NYI
        8 ) search for substrings in spans and link to (a) dictionaries (b) other reports
        9 ) add hyperlinks to substrings
        10 ) create (a) manifest (b) reading order (c) ToC from HTML

        """
        # skip = {"step1"}
        in_dir, session_dir, top_out_dir = self._make_top_in_and_out_and_session()
        # sub_session_list = ["1_4_CMA_3", "5_CMA_3", "6_24_CMA_3"]

        # in_sub_dir = Path(in_dir, session_dir)
        # out_sub_dir = Path(top_out_dir, session_dir)
        force_make_pdf = True  # overrides the "make"
        # file_splitter = "span[@class='Decision']"  # TODO move to dictionary
        # targets = ["decision", "paris", "article", "temperature"]
        # debug = True

        TestUNFCCC.run_pipeline_on_unfccc_session(
            in_dir,
            session_dir,
            top_out_dir=top_out_dir
        )

    def _make_top_in_and_out_and_session(self, in_top=UNFCCC_DIR, out_top=UNFCCC_TEMP_DIR, sub_top="unfcccdocuments1",
                                         session_dir="CMA_3"):
        in_dir = Path(in_top, sub_top)
        top_out_dir = Path(out_top, sub_top)
        return in_dir, session_dir, top_out_dir

    @unittest.skipUnless(AmiAnyTest.run_long() or True, "run occasionally")
    def test_explicit_conversion_pipeline_IMPORTANT_CORPUS(self):
        """reads a corpus of 12 sessions and generates split.html for each
        See test_explicit_conversion_pipeline_IMPORTANT_DEFINITIVE(self): which is run for each session document
        """
        sub_top = "unfcccdocuments1"
        in_dir = Path(UNFCCC_DIR, sub_top)
        top_out_dir = Path(UNFCCC_TEMP_DIR, sub_top)

        session_files = FileLib.posix_glob(str(in_dir) + "/*")
        session_dirs = [d for d in session_files if Path(d).is_dir()]
        print(f">session_dirs {session_dirs}")
        assert len(session_dirs) >= 1

        maxsession = 5  # otherwise runs for ever
        for session_dir in session_dirs[:maxsession]:
            TestUNFCCC.run_pipeline_on_unfccc_session(
                in_dir,
                session_dir,
                top_out_dir=top_out_dir
            )


    def test_create_decision_hyperlink_table(self):
        """creates table of hyperlinks from inline markuo to decisions
        relies on previously created *.marked.html
        """

        def _extract_hyperlinks_to_decisions(marked_file):
            """Currently hyperlinks are
            file:///Users/pm286/workspace/pyamihtml_top/temp/unfccc/unfcccdocuments1//CP_21/Decision_1_CP_21/marked.html
            <a href="../../../../../temp/unfccc/unfcccdocuments1//CMA_3/Decision_1_CMA_3/marked.html">1/CMA.3</a>
            """

            html_elem = lxml.etree.parse(str(marked_file))
            a_elems = html_elem.xpath(".//a[@href[contains(.,'ecision')]]")
            return a_elems

        sub_top = "unfcccdocuments1"
        # in_dir = Path(UNFCCC_TEMP_DIR, sub_top)
        in_dir = Path(Resources.TEST_RESOURCES_DIR, "unfccc", sub_top)
        in_sub_dir = Path(in_dir, "CMA_1")
        insub_sub_dir = Path(in_sub_dir, "Decision_4_CMA_1")
        marked_file = Path(insub_sub_dir, "marked.html")
        assert marked_file.exists()
        marked_elem = HtmlLib.parse_html(marked_file)
        a_elems = _extract_hyperlinks_to_decisions(marked_file)
        assert len(a_elems) > 12

    def test_extract_decision_hyperlinks_from_CORPUS_FAIL(self):
        """iterates over all marked.html and extracts hyperlinks to Decisions
        """
        sub_top = "unfcccdocuments1"
        in_dir = Path(UNFCCC_TEMP_DIR, sub_top)
        outcsv = Path(in_dir, "decision_links.csv")
        outcsv_wt = Path(in_dir, "decision_links_wt.csv")
        TestUNFCCC.create_decision_table(in_dir, outcsv, outcsv_wt=None)

        print(f"wrote csv {outcsv}")

    def test_OBOE_error_for_split_to_marked_FAIL(self):
        """converting a list of split.html to marked.html loses the last element
        tests the pipeline
        """

        session = "CP_21"
        session = "CP_20"

        # infile = Path(UNFCCC_DIR, "unfcccdocuments1", session, "1_CP_21.pdf")
        sub_top = "unfcccdocuments1"
        in_dir = Path(UNFCCC_DIR, sub_top)

        # instem_list = ["1_CP_21", "2_13_CP_21"]
        instem_list = ["1_CP_20", "2_12_CP_20"]

        in_sub_dir = Path(in_dir, session)
        top_out_dir = Path(UNFCCC_TEMP_DIR, sub_top)
        FileLib.force_mkdir(top_out_dir)
        logger.info(f"top_out_dir is {top_out_dir}")
        out_sub_dir = Path(top_out_dir, session)
        logger.info(f"out_sub_dir is {out_sub_dir}")
        FileLib.force_mkdir(out_sub_dir)
        skip_assert = True
        file_splitter = "span[@class='Decision']"  # TODO move to dictionary
        targets = ["decision", "paris"]

        # this needs methods
        class_with_directory_maker = TestUNFCCC
        for instem in instem_list:

            HtmlPipeline.stateless_pipeline(
                file_splitter=file_splitter,
                in_dir=in_dir,
                in_sub_dir=in_sub_dir,
                instem=instem,
                out_sub_dir=out_sub_dir,
                # skip_assert=skip_assert,
                top_out_dir=top_out_dir,
                directory_maker=class_with_directory_maker,
                markup_dict=MARKUP_DICT,
                inline_dict=INLINE_DICT,
                debug=True,
                targets=targets,
            )
        decision = "Decision_1_CP_20"
        split_file = Path(out_sub_dir, decision, "split.html")
        logger.info(f"split file is {split_file}")
        assert split_file.exists()
        marked_file = Path(out_sub_dir, decision, "marked.html")
        logger.info(f"marked file is {marked_file}")
        assert marked_file.exists()

    # def test_subcommands_simple(self):
    #
    #     # run args
    #     AMIClimate().run_command(
    #         ['UNFCCC', '--help'])
    #
    # @unittest.skipUnless(AmiAnyTest.run_long(), "run occasionally")
    # def test_subcommands_long(self):
    #
    #     in_dir, session_dir, top_out_dir = self._make_top_in_and_out_and_session()
    #     try:
    #         AMIClimate().run_command(
    #             ['UNFCCC', '--indir', str(in_dir), '--outdir', str(top_out_dir), '--session', session_dir, '--operation',
    #              UNFCCCArgs.PIPELINE])
    #         raise ValueError("should raise bad argument 'Pipeline' NYI")
    #     except Exception as e:
    #         assert True, "expected fail"






"""
tests networks and graphs
"""
import json
from pathlib import Path
import re

import graphviz
import networkx as nx

from amilib.ami_graph import AmiGraph
from amilib.ami_html import HtmlLib, HtmlEditor
from amilib.file_lib import FileLib
from amilib.util import Util
from amilib.xml_lib import XmlLib
from test.resources import Resources
from test.test_all import AmiAnyTest

logger = Util.get_logger(__name__)
class AmiGraphTest(AmiAnyTest):
    """
    collection of tests for `amilib.graph`
    """

    def test_create_report(self):
        import graphviz

        ipcc = graphviz.Graph('IPCC', filename='fdpclust.gv', engine='fdp')

        glossary = ipcc.node('glossaryx', URL='https://ipcc.ch')
        ipcc.edge('ipcc','glossary')


        with ipcc.subgraph(name='clusterIPCC') as a:
            a.edge('toc1', 'ipcc')
            # {rank = same;
            # wg1.Chap01;
            # wg1.Chap02;
            # wg1.Chap03;
            # wg1.Chap04;
            # wg1.Chap05;
            # wg1.Chap06;}

            with a.subgraph(name='clusterWG1') as wg1:
                wg1.edge('wg1.Chap01', 'toc1')
                wg1.edge('wg1.Chap02', 'toc1')
                wg1.edge('wg1.Chap03', 'toc1')
                wg1.edge('wg1.Chap04', 'toc1')
                wg1.edge('wg1.Chap05', 'toc1')
                wg1.edge('wg1.Chap06', 'toc1')
            a.edge('toc2', 'ipcc')
            with a.subgraph(name='clusterWG2') as wg2:
                wg2.edge('wg2.Chap01', 'toc2')
                wg2.edge('wg2.Chap02', 'toc2')
                wg2.edge('wg2.Chap03', 'toc2')
                wg2.edge('wg2.Chap04', 'toc2')
                wg2.edge('wg2.Chap05', 'toc2')
                wg2.edge('wg2,Chap06', 'toc2')
            a.edge('toc3', 'ipcc')
            with a.subgraph(name='clusterWG3') as wg3:
                wg3.edge('wg3.Chap01', 'toc3')
                wg3.edge('wg3.Chap02', 'toc3')
                wg3.edge('wg3.Chap03', 'toc3')
                wg3.edge('wg3.Chap04', 'toc3')
                wg3.edge('wg3.Chap05', 'toc3')
                wg3.edge('wg3,Chap06', 'toc3')

        ipcc.view()


    def test_extract_toc_graph_from_report_toplevel(self):
        """Sharon please develop for crosschapters"""
        """
        read webpage from IPCC report (WG1/2/3) and extract network of components.
        Will probably not work with actual webpages on web because we need to use headless
        browser. Here we use pre-downloaded page in test/ directory
        """
        editor = HtmlEditor()
        for wg in ["wg1", "wg2", "wg3"]:
            ar6 = Path(Resources.TEST_RESOURCES_DIR, "ar6")
            IN_WG = Path(ar6, wg)
            OUT_WG = Path(Resources.TEMP_DIR, "ipcc", wg)
            inpath = Path(IN_WG, "toplevel.html")

            editor.read_html(inpath)
            json_path = Path(ar6, "edit_toplevel.json")
            logger.info(f"json commands {json_path=}")
            editor.read_commands(json_path)
            editor.execute_commands()
            editor.add_element(parent_xpath="/html/head", tag="style", text="div {border: solid 1px red; margin: 5px;}")

            outpath = Path(OUT_WG, "toplevel.html")
            HtmlLib.write_html_file(editor.html_elem, outpath, debug=True)
            assert outpath.exists()

    def test_strip_single_child_divs(self):
        """
        takes result of applying json commands ("edited_toplevel.html") and
        strips single_child divs
        """
        for wg in ["wg1" , "wg2", "wg3"]:
            inpath = Path(Resources.TEST_RESOURCES_DIR, "ar6", wg, "edited_toplevel.html")
            html_elem = HtmlLib.parse_html(inpath)
            HtmlLib.remove_single_child_divs(html_elem)
            outfile = Path(Resources.TEMP_DIR, "ipcc", wg, "stripped_toplevel.html")
            HtmlLib.write_html_file(html_elem, outfile, debug=True)

    def test_create_toc_tree_graphviz(self):
        """
        wg1 toplevel page with nested divs
        experimental
        """
        for wg in ["wg1", "wg2", "wg3"]:
            print(f"**** current wg {wg}")
            inpath = Path(Resources.TEST_RESOURCES_DIR, "ar6", wg, "stripped_toplevel.html")
            html_elem = HtmlLib.parse_html(inpath)
            ul = AmiGraph.create_nested_uls_from_nested_divs(html_elem)
            outfile = Path(Resources.TEMP_DIR, "ipcc", wg, "toc.html")
            HtmlLib.write_html_file(ul, outfile, debug=True)

    def test_toc_workflow(self):
        """
        creates tocs and trees from toplevel HTML
        collects the components above together
        NEEDS TIDYING
        """
        editor = HtmlEditor()
        for wg in ["wg1", "wg2", "wg3"]:
            print(f"**** current wg {wg}")
            ar6 = Path(Resources.TEST_RESOURCES_DIR, "ar6")
            IN_WG = Path(ar6, wg)
            OUT_WG = Path(Resources.TEMP_DIR, "ipcc", wg)
            inpath = Path(IN_WG, "toplevel.html")

            # apply edit commands
            editor.read_html(inpath)
            json_path = Path(ar6, "edit_toplevel.json")
            logger.info(f"json commands {json_path}")
            editor.read_commands(json_path)
            editor.execute_commands()
            # add box style (not yet implemented in commands
            editor.add_element(parent_xpath="/html/head", tag="style", text="div {border: solid 1px red; margin: 5px;}")

            outpath = Path(OUT_WG, "edited_toplevel.html")
            HtmlLib.write_html_file(editor.html_elem, outpath, debug=True)

            #  read what we have written

            inpath = outpath
            html_elem = HtmlLib.parse_html(inpath)
            HtmlLib.remove_single_child_divs(html_elem)
            outfile = Path(Resources.TEMP_DIR, "ipcc", wg, "stripped_toplevel.html")
            HtmlLib.write_html_file(html_elem, outfile, debug=True)

            # read what we have written
            inpath = outpath
            html_elem = HtmlLib.parse_html(inpath)
            ul = AmiGraph.create_nested_uls_from_nested_divs(html_elem)
            outfile = Path(Resources.TEMP_DIR, "ipcc", wg, "toc.html")
            HtmlLib.write_html_file(ul, outfile, debug=True)
            assert outfile.exists()


    def test_chapter_graph(self):
        """
        simple example of extracting and plotting TOC tree from chapter
        """
        wg = "wg2"
        chapter = "Chapter05"
        # create filename from test/resources
        infile = Path(Resources.TEST_RESOURCES_DIR, "ipcc", "cleaned_content", wg, chapter, "html_with_ids.html")
        gv_output = Path(Resources.TEMP_DIR, "ipcc", wg, chapter, f"toc.gv")
        graph = AmiGraph.create_and_display_chapter_toc_network(wg, chapter, infile, gv_output)
        logger.info(f"wrote graphviz file {gv_output}")
        svg_dir = Path(Resources.TEMP_DIR, "ipcc", wg, chapter)
        svg_output = Path(svg_dir, f"toc.svg")
        graph.format = 'svg'

        # graphviz.render(filepath=svg_output).replace('\\', '/')

        engine = "fdp"
        # engine = "neato"
        svg_path = graph.render(filename=svg_output, engine=engine, format="svg")
        logger.info(f"wrote svg {svg_output}")

# """
# There are two files which are used:-
# 1. TEMP Directory - Resources.TEMP_DIR, "ipcc"
# 2. Original Directory - Resources.TEST_RESOURCES_DIR, "ar6"
#
# Thus, confusion in directory path led to not forming of wg2 and wg3 toc.
# Now, solved by simply changing the directory at required places.
# """

# def test_extract_toc_graph_from_report_toplevel1(self):
#     """
#     read webpage from IPCC report (WG1/2/3) and extract network of components.
#     Will probably not work with actual webpages on web because we need to use headless
#     browser. Here we use pre-downloaded page in test/ directory
#     """
#     editor = HtmlEditor()
#     for wg in ["wg1", "wg2", "wg3"]:
#         ar6 = Path(Resources.TEST_RESOURCES_DIR, "ar6")
#         IN_WG = Path(ar6, wg)
#         OUT_WG = Path(Resources.TEMP_DIR, "ipcc", wg)
#         inpath = Path(IN_WG, "toplevel.html")
#
#         editor.read_html(inpath)
#         json_path = Path(ar6, "edit_toplevel.json")
#         logger.info(f"json commands {json_path}")
#         editor.read_commands(json_path)
#
#         #Checking whether the execute command is working or not
#         try:
#             logger.info(f"Executing commands for {wg} using {json_path}")
#             editor.execute_commands()
#         except Exception as e:
#             logger.error(f"Error while executing commands for {wg}: {e}")
#         editor.add_element(parent_xpath="/html/head", tag="style", text="div {border: solid 1px red; margin: 5px;}")
#
#         #Constructs the output file path inside the TEMP directory
#         outpath = Path(OUT_WG, "toplevel.html")
#         HtmlLib.write_html_file(editor.html_elem, outpath, debug=True)
#
#         #Constructs a path to save the edited HTML in the original Working Group folder
#         edited_path = Path(IN_WG, "edited_toplevel.html")
#         HtmlLib.write_html_file(editor.html_elem, edited_path, debug=True)
#
# def test_strip_single_child_divs1(self):
#     """
#     takes result of applying json commands ("edited_toplevel.html") and
#     strips single_child divs
#     """
#     for wg in ["wg1" , "wg2", "wg3"]:
#
#         """
#         "toplevel.html" and "edited_toplevel.html" is similar, there is no difference.
#         Thus, for classification using a single file, i.e. "edited_toplevel.html"
#         Just there is difference of directory.
#         """
#
#         inpath = Path(Resources.TEST_RESOURCES_DIR, "ar6", wg, "edited_toplevel.html")
#         #Check if the file exists before proceeding
#         if not inpath.exists():
#             logger.error(f"Missing file: {inpath}, skipping {wg}")
#             continue #Skip this WG if the file doesn't exist
#
#         html_elem = HtmlLib.parse_html(inpath)
#         HtmlLib.remove_single_child_divs(html_elem)
#         outfile = Path(Resources.TEMP_DIR, "ipcc", wg, "stripped_toplevel.html")
#         HtmlLib.write_html_file(html_elem, outfile, debug=True)

    def test_junk(self):
        pass

    def test_create_toc_tree_graphviz1(self):
        """
        wg1 toplevel page with nested divs
        experimental
        """
        for wg in ["wg1", "wg2", "wg3"]:
            """ 
                        Taking the output from function 'test_strip_single_child_divs'.
                        The output file is in 'ipcc' directory not in 'ar6'.
                        In this function we are taking the 'ar6' directory, thus creating problem.
                        Changed the directory to the 'ipcc'.
            """

            print(f"**** current wg {wg}")

            #Changing 'Resources.TEST_RESOURCES_DIR' to the 'Resources.TEMP_DIR'
            inpath = Path(Resources.TEMP_DIR, "ipcc", wg, "stripped_toplevel.html")
            html_elem = HtmlLib.parse_html(inpath)
            ul = AmiGraph.create_nested_uls_from_nested_divs(html_elem)
            outfile = Path(Resources.TEMP_DIR, "ipcc", wg, "toc.html")
            HtmlLib.write_html_file(ul, outfile, debug=True)


    def test_toc_workflow1(self):
        """
        creates tocs and trees from toplevel HTML
        adds the components above together
        NEEDS TIDYING
        """
        editor = HtmlEditor()
        networks = {}
        # wgs = ["wg1", "wg2", "wg3"]
        wgs = ["wg1"]

        for wg in wgs:
            logger.info(f"**** current wg {wg}")
            ar6 = Path(Resources.TEST_RESOURCES_DIR, "ar6")
            IN_WG = Path(ar6, wg)
            OUT_WG = Path(Resources.TEMP_DIR, "ipcc", wg)
            inpath = Path(IN_WG, "toplevel.html")

            # apply edit commands
            editor.read_html(inpath)
            json_path = Path(ar6, "edit_toplevel.json")
            logger.info(f"json commands {json_path}")
            editor.read_commands(json_path)
            editor.execute_commands()
            # add box style (not yet implemented in commands
            editor.add_element(parent_xpath="/html/head", tag="style", text="div {border: solid 1px red; margin: 5px;}")

            outpath = Path(OUT_WG, "edited_toplevel.html")
            HtmlLib.write_html_file(editor.html_elem, outpath, debug=True)



            #  read what we have written
            """
            Parsing the edited_toplevel.html to remove the unnecessary tags.
            making it stripped_toplevel.html
            """
            inpath = outpath
            html_elem = HtmlLib.parse_html(inpath)
            HtmlLib.remove_single_child_divs(html_elem)
            outfile = Path(Resources.TEMP_DIR, "ipcc", wg, "stripped_toplevel.html")
            HtmlLib.write_html_file(html_elem, outfile, debug=True)


            # read what we have written
            """
            Taking the 'stripped_toplevel.html' and converting into nested list.
            """
            inpath = outpath
            html_elem = HtmlLib.parse_html(inpath)
            ul = AmiGraph.create_nested_uls_from_nested_divs(html_elem)
            outfile = Path(Resources.TEMP_DIR, "ipcc", wg, "toc.html")
            HtmlLib.write_html_file(ul, outfile, debug=True)


            """
            Taking the toc.html and creating graph by the function - 'create_graph_from_toc' that uses "Networkx".
            """
            # Parse the toc.html and create a graph
            toc_path = Path(Resources.TEMP_DIR, "ipcc", wg, "toc.html")
            ami_graph = AmiGraph()
            network = ami_graph.create_network_from_toc(toc_path)


            """
            Made a dictionary to take different working group (wg1, wg2, wg3)
            """
            networks[wg] = network


            """
            Visualization of the network after creating it.
            Saved it in png format.
            """
            #Visualize the network for each wg using Networkx
            for wg, network in networks.items():
                print(f"Visualizing graphs for {wg}")
                AmiGraph.visualize_graph_with_networkx(network, title = f"Network for {wg}", wg = wg)





    def test_graphviz2(self):
        """
        probably a duplicate
        """
        import graphviz
        gv_output = 'fdpclust.gv'
        ipcc = graphviz.Graph('IPCC', filename=gv_output, engine='fdp')

        # glossary = ipcc.node('glossaryx', URL='https://ipcc.ch')
        ipcc.edge('ipcc', 'glossary')

        with ipcc.subgraph(name='clusterIPCC') as a:
            a.edge('toc1', 'ipcc')

            with a.subgraph(name='clusterWG1') as wg1:
                wg1.edge('wg1.Chap01', 'toc1')
                wg1.edge('wg1.Chap02', 'toc1')
                wg1.edge('wg1.Chap03', 'toc1')
                wg1.edge('wg1.Chap04', 'toc1')
                wg1.edge('wg1.Chap05', 'toc1')
                wg1.edge('wg1.Chap06', 'toc1')
            a.edge('toc2', 'ipcc')
            with a.subgraph(name='clusterWG2') as wg2:
                wg2.edge('wg2.Chap01', 'toc2')
                wg2.edge('wg2.Chap02', 'toc2')
                wg2.edge('wg2.Chap03', 'toc2')
                wg2.edge('wg2.Chap04', 'toc2')
                wg2.edge('wg2.Chap05', 'toc2')
                wg2.edge('wg2,Chap06', 'toc2')
            a.edge('toc3', 'ipcc')
            with a.subgraph(name='clusterWG3') as wg3:
                wg3.edge('wg3.Chap01', 'toc3')
                wg3.edge('wg3.Chap02', 'toc3')
                wg3.edge('wg3.Chap03', 'toc3')
                wg3.edge('wg3.Chap04', 'toc3')
                wg3.edge('wg3.Chap05', 'toc3')
                wg3.edge('wg3,Chap06', 'toc3')

        ipcc.view()
